/*
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Single raw character reading routines.
 * This inputs single characters without echoing (assuming the original
 * dpyInit modes specified no echo and cbreak or raw modes).
 * This has special handling for signals so the read calls will return
 * reliably indicating interruptions (especially to handle screen resizes).
 */

#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <poll.h>

#include "dpy_priv.h"


/*
 * Saved character that was previously read, or -1 if none.
 */
static int savedChar = -1;

/*
 * File handles for the reading and writing ends of a pipeline.
 */
static int pipeFileHandles[2];
static BOOL isPipeOpen;


/*
 * Local routines.
 */
static int dpyDoRead(void);
static BOOL setNonBlockingMode(int);


/*
 * Test to see if input is ready from STDIN.
 * Returns 1 if input is ready, 0 if input is not ready, or DPY_ERROR on error.
 * This does not check whether or not DPY_INTR should be returned.
 * NOTE: This may actually read a character and save it away for later use.
 * Therefore the only valid way to read a character from STDIN after this
 * routine says a character is ready is to use dpyInputChar (or dpyRead
 * which calls dpyInputChar).
 */
int
dpyTestInput(void)
{
#ifdef FIONREAD
	int n;

	if (savedChar >= 0)
		return 1;

	if (ioctl(STDIN, FIONREAD, &n) < 0)
		return DPY_ERROR;

	return (n > 0);
#else
	int flags;
	int n;
	unsigned char ch;

	if (savedChar >= 0)
		return 1;

	flags = fcntl(STDIN, F_GETFL, 0);

	if (flags == -1)
		return DPY_ERROR;

	if ((flags & O_NONBLOCK) == 0)
	{
		if (fcntl(STDIN, F_SETFL, flags | O_NONBLOCK) < 0)
			return DPY_ERROR;
	}

	n = read(STDIN, &ch, 1);

	if ((flags & O_NONBLOCK) == 0)
	{
		if (fcntl(STDIN, F_SETFL, flags) < 0)
			return DPY_ERROR;
	}

	if (n < 0)
		return DPY_ERROR;

	if (n > 0)
		savedChar = ch;

	return n;
#endif
}


/*
 * Read a character from STDIN, blocking if necessary.
 * This may return a character saved from a previous dpyTestInput call.
 * Returns the character read on success, or a negative value on an error
 * (e.g., DPY_EOF, DPY_INTR, or DPY_ERROR).
 */
int
dpyInputChar(void)
{
	int status;

	if (savedChar >= 0)
	{
		int ch = savedChar;
		savedChar = -1;

		return ch;
	}

	/*
	 * Wait indefinitely for a character from the terminal
	 * or for an interruption by a signal.
	 */
	status = dpyPollInput(-1);

	if (status < 0)
		return status;

	/*
	 * If a character was ready to read then return it.
	 */
	if (status > 0)
		return dpyDoRead();

	/*
	 * Poll misbehaved somehow.
	 * It should not have returned early since we had no timeout.
	 */
	return DPY_ERROR;
}


/*
 * Perform a poll to wait the specified time (in milliseconds)
 * either for data to arrive on STDIN or for a signal to be delivered
 * and stored in our pipeline.  A negative time specifies waiting
 * indefinitely.  Returns 0 if the timeout occurred, 1 to indicate a
 * character is ready to read from STDIN, or a negative value on an
 * error (e.g., DPY_INTR, DPY_EOF, or DPY_ERROR).
 */
int
dpyPollInput(int msTimeout)
{
	struct pollfd pollTable[2];
	int count;

	/*
	 * Make sure the pipeline has been opened.
	 */
	if (!dpyOpenPipeline())
		return FALSE;

	/*
	 * Fill out the poll table with entries for STDIN and the
	 * read side of the pipeline.
	 */
	pollTable[0].fd = STDIN;
	pollTable[0].events = POLLIN;
	pollTable[0].revents = 0;

	pollTable[1].fd = pipeFileHandles[0];
	pollTable[1].events = POLLIN;
	pollTable[1].revents = 0;

	/*
	 * Wait for data to arrive or for the timeout.
	 */
	count = poll(pollTable, 2, msTimeout);

	if (count < 0)
	{
		/*
		 * If poll failed but it wasn't interrupted then
		 * the problem is serious.
		 */
		if (errno != EINTR)
			return DPY_ERROR;

		/*
		 * The poll was interrupted due to a signal (e.g., terminal
		 * resizing).  Clear the pipeline since it might have data
		 * stored from the signal handler, and there is no need to
		 * return an additional DPY_INTR value later.
		 */
		dpyClearInterrupt();

		return DPY_INTR;
	}

	/*
	 * If nothing is ready to read then the timeout occurred.
	 */
	if (count == 0)
		return 0;

	/*
	 * If STDIN has data then return that fact.
	 */
	if (pollTable[0].revents & POLLIN)
		return 1;

	/*
	 * If the pipeline has data then act like we were interrupted.
	 * First read the data in the pipeline so that it is ready
	 * for a new interrupt.
	 */
	if (pollTable[1].revents & POLLIN)
	{
		dpyClearInterrupt();

		return DPY_INTR;
	}

	/*
	 * We shouldn't get here since poll indicated one of the
	 * file descriptors was ready.
	 */
	return DPY_ERROR;
}


/*
 * Read a character from STDIN, blocking if necessary.
 * Returns the character on success or a negative value on an error
 * (e.g., DPY_INTR, DPY_EOF, or DPY_ERROR).
 */
static int
dpyDoRead(void)
{
	int n;
	unsigned char ch;

	n = read(STDIN, &ch, 1);

	if (n < 0)
	{
		if (errno == EINTR)
			return DPY_INTR;

		return DPY_ERROR;
	}

	if (n == 0)
		return DPY_EOF;

	return ch;
}


/*
 * Return the file handle for the reading side of our pipeline.
 * This can be used by the programs's own poll logic to wait on.
 * The file handle should not be modified, closed, or read from.
 * Returns the file handle, or -1 if the pipe could not be opened.
 */
int
dpyGetPollFileHandle(void)
{
	if (!dpyOpenPipeline())
		return -1;

	return pipeFileHandles[0];
}


/*
 * Try to open our pipeline that is used to remember resize signals.
 * Nothing is done if the pipeline is already opened.
 * Returns TRUE on success, or FALSE on an error.
 */
BOOL
dpyOpenPipeline(void)
{
	if (isPipeOpen)
		return TRUE;

	/*
	 * The pipeline is not yet open, so allocate the file handles.
	 */
	pipeFileHandles[0] = -1;
	pipeFileHandles[1] = -1;

	if (pipe(pipeFileHandles) < 0)
		return FALSE;

	isPipeOpen = TRUE;

	/*
	 * Set the pipe file handles to non-blocking mode.
	 */
	if (!setNonBlockingMode(pipeFileHandles[0]))
		return FALSE;

	if (!setNonBlockingMode(pipeFileHandles[1]))
		return FALSE;

	return TRUE;
}


/*
 * Set the file descriptor to be in non-blocking mode.
 * Returns TRUE on success, or FALSE on failure.
 */
static BOOL
setNonBlockingMode(int fd)
{
	int flags = fcntl(fd, F_GETFL, 0);

	if (flags == -1)
		return FALSE;

	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0)
		return FALSE;

	return TRUE;
}


/*
 * Try to close our pipeline to free its file descriptors.
 * Nothing is done if the pipeline had not been opened.
 * Returns TRUE on success, or FALSE on an error.
 */
BOOL
dpyClosePipeline(void)
{
	BOOL isSuccess = TRUE;

	if (!isPipeOpen)
		return TRUE;

	isPipeOpen = FALSE;

	if (close(pipeFileHandles[0]) < 0)
		isSuccess = FALSE;

	if (close(pipeFileHandles[1]) < 0)
		isSuccess = FALSE;

	pipeFileHandles[0] = -1;
	pipeFileHandles[1] = -1;

	return isSuccess;
}


/*
 * Set up so that a read from dpyInputChar will not block, but will
 * instead return DPY_INTR if no actual character is available from STDIN.
 * WARNING: This can be called from a signal handler and so we can't do
 * anything complicated.  But we CAN set a flag and write a byte of data
 * into our pipeline so that dpyPollInput will later return DPY_INTR.
 * The write call won't block since we have set non-blocking mode.
 * Returns TRUE on success, or FALSE on an error.
 */
BOOL
dpySetInterrupt(void)
{
	if (!dpyOpenPipeline())
		return FALSE;

	if (write(pipeFileHandles[1], "", 1) < 0)
		return (errno == EAGAIN);

	return TRUE;
}


/*
 * Clear the data in the pipeline by reading it all.
 * The read calls won't block since we have set non-blocking mode.
 * Returns TRUE on success, or FALSE on an error.
 */
BOOL
dpyClearInterrupt(void)
{
	if (!isPipeOpen)
		return TRUE;

	while (TRUE)
	{
		char buf[32];

		int count = read(pipeFileHandles[0], buf, sizeof(buf));

		if (count > 0)
			continue;

		return ((count == 0) || (errno == EAGAIN));
	}
}

/* END CODE */
