/*
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include <memory.h>
#include <termcap.h>

#include "dpy_priv.h"


#define	CLEARSIZE	5	/* number of spaces worth clear eol action */
#define	MOVESIZE	8	/* characters worth cursor addressing action */


/*
 * Local routines.
 */
static	void dpyCursor(void);
static	int dpyTPutS(int);


/*
 * Make the screen match the data as previously written by the user.
 * This is what does the actual output to the terminal by only updating
 * the changed text.  But if the dpySetRedraw routine had been called,
 * then we do a dpyRefresh instead.
 */
void
dpyUpdate(void)
{
	WINDOW * wp = _dpy_window_;

	char * cp;		/* current character */
	char * screenCp;	/* screen character pointer */
	char * spaceCp;		/* cp where spaces remain in row */
	char * endRow;		/* end of screen row */
	char * begRow;		/* beginning of screen row */
	int row;		/* current row number */

	/*
	 * Disable stopping while we do the update for safety.
	 * If stopping was requested we will do it when we return.
	 */
	wp->isStopSafe = FALSE;

	/*
	 * Do a full redraw instead if the program had requested it.
	 */
	if (wp->isRedrawNeeded)
	{
		dpyRedraw();

		return;
	}

	/*
	 * If this is the first output then clear the screen.
	 */
	if (!wp->isDidOutput)
	{
		wp->isDidOutput = TRUE;

		fwrite(wp->tc_ho, 1, wp->tc_hocc, stdout);
		fwrite(wp->tc_cd, 1, wp->tc_cdcc, stdout);
	}

	/*
	 * Start the loop from the beginning of the changed region.
	 */
	cp = wp->begChange;
	screenCp = wp->screen + (cp - wp->begData);
	endRow = cp;
	begRow = 0;
	row = 0;

	/*
	 * Keep looping looking for the next difference until we
	 * reach the end of the changed region.
	 */
	while (cp < wp->endChange)
	{
		int sameCount = dpyStrDif(cp, screenCp, wp->endChange - cp);

		cp += sameCount;
		screenCp += sameCount;

		if (cp >= wp->endChange)
			break;

		spaceCp = 0;

		/*
		 * If we are moving past the current line then recalculate
		 * the screen row information for the new position and see
		 * how many spaces are at the end of the row.
		 */
		if (cp >= endRow)
		{
			row = (cp - wp->begData) / wp->rowDelta;
			begRow = wp->begData + (row * wp->rowDelta);
			endRow = begRow + wp->colCount;
			spaceCp = endRow - 1;

			while ((spaceCp >= begRow) && (*spaceCp == SPACE))
				spaceCp--;

			spaceCp++;
		}

		/*
		 * Move to the position of this character if necessary.
		 */
		dpyDoMove(row, cp - begRow, begRow);

		/*
		 * Clear the rest of the line using a terminal sequence
		 * if there are enough spaces to make it worth it.
		 */
		if (spaceCp && (cp >= spaceCp + CLEARSIZE))
		{
			fwrite(wp->tc_ce, 1, wp->tc_cecc, stdout);

			while (cp < endRow)
			{
				*screenCp++ = SPACE;
				cp++;
			}

			continue;
		}

		/*
		 * Update the character at this position.
		 */
		putchar(*cp);
		*screenCp++ = *cp++;

		/*
		 * Fix up last column position based on the terminal type,
		 * either not advancing it or else moving to the next line.
		 */
		if (++wp->termCol >= wp->colCount)
		{
			wp->termCol--;

			if (wp->tc_am)
			{
				wp->termRow++;
				wp->termCol = 0;
			}
		}
	}

	/*
	 * Set the changed region to the reversed order so that
	 * no change region is in effect until something is written.
	 */
	wp->begChange = wp->endData;
	wp->endChange = wp->begData;

	/*
	 * Position the cursor to the current position if enabled
	 * and make sure all the output is flushed.
	 */
	if (!wp->isLeaveCursor)
		dpyCursor();

	fflush(stdout);

	/*
	 * Finally if we had a signal to stop the program while we
	 * were updating then do it now that it is safe.
	 */
	wp->isStopSafe = TRUE;

	if (wp->isStopPending)
		dpyStop();
}


/*
 * Set the terminal cursor at the current write location.
 * If the window is full, the cursor is placed at the front of the
 * last line in the window.  If lines are not being wrapped and the
 * line is full, the cursor is placed at the end of the line.
 * Otherwise, the cursor is placed at the location being written to next.
 */
static void
dpyCursor(void)
{
	WINDOW * wp = _dpy_window_;

	char * cp = wp->cp;	/* current write location */
	char * begRow;		/* beginning of current row */
	int row;		/* row number */

	if (wp->isFull)
	{
		cp = wp->endWin;
	}
	else if (cp >= wp->endRow)
	{
		if (wp->isNoLineWrap || (wp->begRow >= wp->endWin))
			cp = wp->endRow - 1;
		else
			cp = wp->begRow + wp->rowDelta;
	}

	row = (cp - wp->begData) / wp->rowDelta;
	begRow = wp->begData + (row * wp->rowDelta);

	dpyDoMove(row, cp - begRow, begRow);
}


/*
 * Subroutine to move to the given location on the screen.  The third argument
 * is a pointer to beginning of the desired row in case we find it is faster
 * to type the intervening characters.  If NULL, we must use addressing.
 */
void
dpyDoMove(int row, int col, const char * cp)
{
	WINDOW * wp = _dpy_window_;

	/*
	 * If we are already at the required position then do nothing.
	 */
	if ((wp->termRow == row) && (wp->termCol == col))
		return;

	/*
	 * See if we are moving just a few characters ahead on the line.
	 * If so then just print the intervening characters to get there.
	 */
	if (cp && (row == wp->termRow) && (col >= wp->termCol)
		&& (col < wp->termCol + MOVESIZE))
	{
		cp += wp->termCol;

		while (wp->termCol < col)
		{
			putchar(*cp);
			cp++;
			wp->termCol++;
		}

		return;
	}

	/*
	 * Check for moving to the next row.
	 */
	if ((col == 0) && (row == wp->termRow + 1))
	{
		if (wp->isRaw)
			putchar(RET);

		putchar(EOL);
		wp->termRow++;
		wp->termCol = 0;

		return;
	}

	/*
	 * Arbitrary movement, so use cursor addressing.
	 */
	tputs(tgoto(wp->tc_cm, col, row), 1, dpyTPutS);

	wp->termRow = row;
	wp->termCol = col;
}


/*
 * Local routine called by tputs to print a character.
 */
static int
dpyTPutS(int ch)
{
	putchar(ch);

	return 0;
}


/*
 * Redraw the screen to fix glitches.
 * This sets the screen image to spaces, marks the whole screen as
 * being changed, clears the terminal, and then does an update.
 */
void
dpyRedraw(void)
{
	WINDOW * wp = _dpy_window_;

	memset(wp->screen, SPACE, wp->rowCount * wp->rowDelta);

	wp->termRow = 0;
	wp->termCol = 0;
	wp->begChange = wp->begData;
	wp->endChange = wp->endData;

	wp->isStopSafe = FALSE;
	wp->isRedrawNeeded = FALSE;
	wp->isDidOutput = TRUE;

	fwrite(wp->tc_ho, 1, wp->tc_hocc, stdout);
	fwrite(wp->tc_cd, 1, wp->tc_cdcc, stdout);

	dpyUpdate();
}


/*
 * Set that the next dpyUpdate call will do a dpyRedraw instead.
 */
void
dpySetRedraw(void)
{
	WINDOW * wp = _dpy_window_;

	wp->isRedrawNeeded = TRUE;
}

/* END CODE */
