/*
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include "dpy_priv.h"


#define	ROWCHAR		'-'
#define	COLCHAR		'|'
#define	CORNERCHAR	'+'


static void storeChar(int, int, int);


/*
 * Set the row and column boundaries of the current window.
 * Negative numbers indicate backwards from last row or column.
 * The write location is set to the top left of the window.
 * Returns TRUE on success or FALSE if arguments are out of bounds.
 */
BOOL
dpyWindow(int minRow, int maxRow, int minCol, int maxCol)
{
	WINDOW * wp = _dpy_window_;

	if (minRow < 0)
		minRow += wp->rowCount;

	if (maxRow < 0)
		maxRow += wp->rowCount;

	if (minCol < 0)
		minCol += wp->colCount;

	if (maxCol < 0)
		maxCol += wp->colCount;

	/*
	 * Check for illegal size.
	 */
	if (((unsigned) minRow > maxRow)
		|| ((unsigned) maxRow >= wp->rowCount)
		|| ((unsigned) minCol > maxCol)
		|| (maxCol >= wp->colCount))
	{
		return FALSE;
	}

	wp->tabColumn = 0;
	wp->begWin = wp->begData + (minRow * wp->rowDelta) + minCol;
	wp->endWin = wp->begWin + ((maxRow - minRow) * wp->rowDelta);
	wp->begRow = wp->begWin;
	wp->endRow = wp->begRow + (maxCol - minCol + 1);
	wp->cp = wp->begRow;

	/*
	 * Reset the flags and tab size to the defined defaults.
	 */
	dpySetFlags(-1, wp->defaultFlags);
	dpySetTabSize(wp->defaultTabSize);

	wp->isFull = FALSE;

	return TRUE;
}


/*
 * Set the row and column boundaries of the current window and draw a
 * rectangular frame around that area.  Output is set to only use the
 * inner part of the frame (two lines and two columns smaller than the
 * specified size).  Negative numbers indicate backwards from the last
 * row or column. If this call is made multiple times then the frame
 * edges can overlap if desired.  The write location is set to the top
 * left of the inside of the frame.  Returns TRUE on success or FALSE
 * if arguments are out of bounds.
 */
BOOL
dpyFrame(int minRow, int maxRow, int minCol, int maxCol)
{
	WINDOW * wp = _dpy_window_;
	int i;

	if (minRow < 0)
		minRow += wp->rowCount;

	if (maxRow < 0)
		maxRow += wp->rowCount;

	if (minCol < 0)
		minCol += wp->colCount;

	if (maxCol < 0)
		maxCol += wp->colCount;

	/*
	 * Check for illegal size.
	 */
	if (((unsigned) minRow > maxRow)
		|| ((unsigned) maxRow >= wp->rowCount)
		|| ((unsigned) minCol > maxCol)
		|| (maxCol >= wp->colCount))
	{
		return FALSE;
	}

	/*
	 * Make sure that the frame doesn't exhaust the window,.
	 */
	if ((minRow + 2 > maxRow) || (minCol + 2 > maxCol))
		return FALSE;

	/*
	 * Draw the frame.
	 */
	storeChar(minRow, minCol, CORNERCHAR);
	storeChar(minRow, maxCol, CORNERCHAR);
	storeChar(maxRow, minCol, CORNERCHAR);
	storeChar(maxRow, maxCol, CORNERCHAR);

	for (i = minCol + 1; i < maxCol; i++)
	{
		storeChar(maxRow, i, ROWCHAR);
		storeChar(minRow, i, ROWCHAR);
	}

	for (i = minRow + 1; i < maxRow; i++)
	{
		storeChar(i, minCol, COLCHAR);
		storeChar(i, maxCol, COLCHAR);
	}

	/*
	 * Now set the inner window as the one to draw in.
	 */
	return dpyWindow(minRow + 1, maxRow - 1, minCol + 1, maxCol - 1);
}


/*
 * Store a character into the screen at the specified position.
 * The coordinates are the raw screen coordinates, not the current window
 * coordinates.  No validation is done since this is only used locally.
 */
static void
storeChar(int screenRow, int screenCol, int ch)
{
	WINDOW * wp = _dpy_window_;

	char * cp = wp->begData + (screenRow * wp->rowDelta) + screenCol;

	if (*cp == ch)
		return;

	dpyMarkChange(cp);

	*cp++ = ch;

	dpyMarkChange(cp);
}

/* END CODE */
